#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda_runtime.h>

// Maximum number of floats that can be stored in constant memory (16384 floats ~ 64KB)
#define MAX_CLUSTERS_SIZE 16384

// Declare constant memory for the clusters tensor (expected shape [D, K+G])
__constant__ float d_clusters[MAX_CLUSTERS_SIZE];

// Modular device function for warp-level reduction
__device__ float warpReduceSum(float val) {
    // Use full warp mask
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        val += __shfl_down_sync(0xFFFFFFFF, val, offset);
    }
    return val;
}

// Modular device function to compute partial dot product for one row and one column
// Each thread in a warp computes a partial sum over elements of D using a stride of warpSize
__device__ float dotProduct(const float* __restrict__ x, int row, int D, int N_out, int col, int lane) {
    float sum = 0.0f;
    for (int k = lane; k < D; k += warpSize) {
        // Access clusters from constant memory; assumed layout: [D, N_out] (column-major for clusters)
        sum += x[row * D + k] * d_clusters[k * N_out + col];
    }
    return sum;
}

// Modular warp-level kernel for matrix multiplication using the above device functions
// Computes out = x * clusters, where x is [M, D] and clusters (in constant memory) is [D, N_out].
// Each warp computes one output element (dot product) by having its lanes cooperate.
__global__ void matmul_warp_kernel(const float* __restrict__ x, float* __restrict__ out, int M, int D, int N_out) {
    // Determine the warp id and lane id within the warp
    int warp_id = (blockIdx.x * blockDim.x + threadIdx.x) / warpSize;
    int lane = threadIdx.x & (warpSize - 1);

    // Each warp is responsible for one element of the output matrix
    if (warp_id < M * N_out) {
        int row = warp_id / N_out;
        int col = warp_id % N_out;
        
        // Compute partial dot product using the modular dotProduct function
        float partial = dotProduct(x, row, D, N_out, col, lane);
        
        // Reduce the partial sums across the warp
        float sum = warpReduceSum(partial);
        
        // Lane 0 writes the final result
        if (lane == 0) {
            out[row * N_out + col] = sum;
        }
    }
}

// Forward function for NetVLAD with ghost clusters using modular CUDA device functions
// This function first computes the assignment via custom matrix multiplication (with warp-level reduction),
// then applies batch normalization, softmax, VLAD aggregation, and normalization as in the original implementation.

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor clusters,
    torch::Tensor clusters2,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_mean,
    torch::Tensor bn_var,
    bool is_training,
    int64_t cluster_size,
    int64_t feature_size
) {
    // x is expected to be [B, N, D]
    auto B = x.size(0);
    auto N = x.size(1);
    auto D = x.size(2);
    TORCH_CHECK(D == feature_size, "feature_size mismatch.");

    // Flatten x to [M, D] where M = B * N
    int64_t M = B * N;
    auto x_flat = x.reshape({M, D}).contiguous();
    
    // clusters is [D, (K+G)] where (K+G) >= cluster_size
    int N_out = clusters.size(1);
    auto assignment = torch::empty({M, N_out}, x.options());

    // Use the custom warp-level kernel if clusters fits in constant memory
    if (clusters.numel() <= MAX_CLUSTERS_SIZE) {
        auto clusters_contig = clusters.contiguous();
        size_t clusters_bytes = clusters_contig.numel() * sizeof(float);
        cudaError_t err = cudaMemcpyToSymbol(d_clusters, clusters_contig.data_ptr<float>(), clusters_bytes);
        TORCH_CHECK(err == cudaSuccess, "cudaMemcpyToSymbol failed: ", cudaGetErrorString(err));

        // Total output elements = M * N_out, with each warp computing one element
        int total_warps = M * N_out;
        int threadsPerBlock = 256; // e.g., 8 warps per block
        int warpsPerBlock = threadsPerBlock / 32;
        int blocks = (total_warps + warpsPerBlock - 1) / warpsPerBlock;

        matmul_warp_kernel<<<blocks, threadsPerBlock>>>(
            x_flat.data_ptr<float>(),
            assignment.data_ptr<float>(),
            M, D, N_out
        );
        err = cudaGetLastError();
        TORCH_CHECK(err == cudaSuccess, "Kernel launch failed: ", cudaGetErrorString(err));
        err = cudaDeviceSynchronize();
        TORCH_CHECK(err == cudaSuccess, "Kernel sync failed: ", cudaGetErrorString(err));
    } else {
        // Fallback to regular matmul if clusters does not fit in constant memory
        assignment = at::matmul(x_flat, clusters);
    }

    // Batch normalization
    assignment = at::batch_norm(
        assignment, bn_weight, bn_bias, bn_mean, bn_var,
        is_training, 0.1, 1e-5, true
    );

    // Softmax along dimension 1 and removing ghost clusters to yield [M, cluster_size]
    assignment = at::softmax(assignment, 1);
    assignment = assignment.narrow(1, 0, cluster_size);

    // Reshape to [B, N, cluster_size]
    assignment = assignment.reshape({B, N, cluster_size});

    // Compute a_sum = sum across the N dimension -> [B, 1, cluster_size]
    auto a_sum = assignment.sum(1, /*keepdim=*/true);
    // Multiply by clusters2 (expected shape: [1, D, cluster_size])
    auto a = a_sum * clusters2;

    // Transpose assignment to [B, cluster_size, N]
    assignment = assignment.transpose(1, 2);
    // Reshape x back to [B, N, D]
    auto x_reshaped = x_flat.reshape({B, N, D});

    // VLAD aggregation: compute [B, cluster_size, D] via bmm and then transpose to [B, D, cluster_size]
    auto vlad = at::bmm(assignment, x_reshaped);
    vlad = vlad.transpose(1, 2);
    vlad = vlad - a;

    // Intra-normalize across the D dimension
    vlad = vlad / (vlad.norm(2, {1}, /*keepdim=*/true) + 1e-12);
    
    // Flatten to [B, D * cluster_size] and perform final L2 normalization
    vlad = vlad.reshape({B, D * cluster_size});
    vlad = vlad / (vlad.norm(2, {1}, /*keepdim=*/true) + 1e-12);

    return vlad;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "NetVLAD with ghost clusters (Modular CUDA device functions)");
}
